package net.es.oscars.policy.idc;

import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.Arrays;
import java.util.List;

import org.apache.log4j.Logger;

import net.es.oscars.policy.config.ConfigUtil;
import net.es.oscars.policy.config.PolicyParserException;
import net.es.oscars.policy.config.ParserUtil;
import net.es.oscars.policy.config.RuleSetParser;
import net.es.oscars.policy.config.VarParser;
import net.es.oscars.policy.database.LTypeDAO;
import net.es.oscars.policy.database.OperatorDAO;
import net.es.oscars.policy.database.Rule;
import net.es.oscars.policy.database.RuleDAO;

public class IDCVarParser implements VarParser{
    private Logger log;
    private String name;
    private ConfigUtil configUtil;
    private RuleDAO ruleDAO;
    private LTypeDAO ltypeDAO;
    private OperatorDAO operatorDAO;
    
    static final public List<String> specialFields = Arrays.asList(new String[]{
            "user", "status"});
    static final public List<String> specialOps = Arrays.asList(new String[]{
            "=="});
    static final public List<String> outputTypes = Arrays.asList(new String[]{
            "sum"});
    
    public IDCVarParser(){
        this.log = Logger.getLogger(this.getClass());
    }
    
    public void init(ConfigUtil configUtil, RuleDAO ruleDAO,
            LTypeDAO ltypeDAO,  OperatorDAO operatorDAO){
        this.configUtil = configUtil;
        this.ruleDAO = ruleDAO;
        this.ltypeDAO = ltypeDAO;
        this.operatorDAO = operatorDAO;
    }
    
    public void parseRule(StreamTokenizer tokenizer, List<Rule> ruleStack,
            String token, String normalizedToken, int tokenType, String ruleOp)
            throws IOException, PolicyParserException {
        
        String lType = null;
        String rType = null;
        String lValue = null;
        List<String> ops = null;
        
        if(specialFields.contains(normalizedToken)){
            lType = "XPATH_STRING";
            rType = "STRING";
            lValue = this.getXPath(normalizedToken);
            ops = specialOps;
        }else if(this.configUtil.getStringFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? RuleSetParser.tokenTypes[tokenType-1] : "XPATH_NODE";
            rType = "STRING";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = RuleSetParser.stringOps;
        }else if(this.configUtil.getNumberFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? RuleSetParser.tokenTypes[tokenType-1] : "XPATH_NUMBER";
            rType = "NUMBER";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = RuleSetParser.numberOps;
        }else if(this.configUtil.getBooleanFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? RuleSetParser.tokenTypes[tokenType-1] : "XPATH_BOOLEAN";
            rType = "BOOLEAN";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = RuleSetParser.booleanOps;
        }else if(this.configUtil.getDateFields().contains(normalizedToken)){
            lType = (tokenType > ParserUtil.POSITION_TOKEN) ? RuleSetParser.tokenTypes[tokenType-1] : "XPATH_NUMBER";
            rType = "DATE";
            lValue = this.prepareXPath(token, normalizedToken, tokenType);
            ops = RuleSetParser.dateOps;
        }else{
            throw new PolicyParserException("Invalid token '" + token + "' at line " + tokenizer.lineno());
        }
        
        tokenizer.nextToken();
        this.log.info("parseRule." + lType + ".start");
        String op = ParserUtil.readToken(tokenizer);
        if(op == null || (!ops.contains(op))){
            throw new PolicyParserException("Invalid operator '" + op + "' at line " + tokenizer.lineno());
        }
        tokenizer.nextToken();
        String value = ParserUtil.readToken(tokenizer);
        //validate date format
        if("DATE".equals(rType) && (!this.configUtil.validateDateField(value))){
            throw new PolicyParserException("Invalid date '" + value + "' at line " + tokenizer.lineno());
        }
        //check if special variable
        if(lValue.matches("#.+#")){
            lType = "SPECIAL_VAR";
            lValue = lValue.replaceAll("#","");
        }
        tokenizer.nextToken();
        if(!";".equals(ParserUtil.readToken(tokenizer))){
            throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
        }
        
        Rule rule = new Rule();
        
        rule.setLType(ltypeDAO.queryByParam("name", lType));
        rule.setLValue(lValue);
        rule.setOperator(operatorDAO.findByTypeName(rType, op));
        rule.setRValue(value);
        Rule existingRule = ruleDAO.findMatchingRule(rule);
        if(existingRule != null){
            rule = existingRule;
        }else{
            ruleDAO.create(rule);
        }
        
        //look at last rule
        if(ruleStack.get(0) == null){
            ruleStack.remove(0);
            ruleStack.add(0, rule);
        }else {
            Rule combinedRule = ParserUtil.combineRule(ruleStack.get(0), rule, 
                    ruleOp, this.ltypeDAO, this.operatorDAO, this.ruleDAO);
            ruleStack.remove(0);
            ruleStack.add(0, combinedRule);
        }
    }
    
    public void setName(String name) {
        this.name = name;
    }
    
    public String getName() {
        return this.name;
    }
    
    public void destroy() {
        // TODO Auto-generated method stub
        
    }
    
    private String prepareXPath(String token, String normalizedToken, int tokenType) {
        String xPath = this.configUtil.getXPathMap().get(normalizedToken);
        if(tokenType == ParserUtil.POSITION_TOKEN){
            String[] parts = token.split("-");
            xPath = xPath.replace("#N#", parts[parts.length-1]);
        }
        
        //Schema version specific
        xPath = xPath.replaceAll("reservationResource", "resDetails");
        
        return xPath;
    }

    public boolean isValidOutputType(String outputType) {
        return outputTypes.contains(outputType);
    }

    public String getOpType(String outputType, List<String> validOps) throws PolicyParserException {
        String rType = "";
        if("sum".equals(outputType)){
            rType = "NUMBER";
            validOps.addAll(RuleSetParser.numberOps);
        }else{
            throw new PolicyParserException("Invalid output type '" + outputType +"'");
        }
        
        return rType;
    }
    
    public String getXPath(String type) throws PolicyParserException{
        if("status".equals(type)){
            return IDCPolicyExt.STATUS_XPATH;
        }else if("user".equals(type)){
            return IDCPolicyExt.USER_XPATH;
        }else{
            throw new PolicyParserException("Cannot map '"+type+"' to XPath");
        }
    }

}
